///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: _xsqlda.cpp,v 1.26.2.2 2004/08/11 14:13:37 epocman Exp $
//	Subject : IBPP, XSDA class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* XSQLDA == See Interbase 6.0 C-API
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

#include <limits>

#include <math.h>
#include <time.h>

using namespace ibpp_internals;

namespace ibpp_internals
{
	const double dscales[19] =
	{
		1, 1E1, 1E2, 1E3, 1E4,	1E5, 1E6, 1E7, 1E8,
		1E9, 1E10, 1E11, 1E12, 1E13, 1E14, 1E15,
		1E16, 1E17, 1E18
	};

// Many compilers confuses those following min/max with macros min and max !
#undef min
#undef max

#ifdef IBPP_DMC // Needs to break-down the declaration else compiler crash (!)
	const std::numeric_limits<short> short_limits;
	const std::numeric_limits<long> long_limits;
	const short minshort = short_limits.min();
	const short maxshort = short_limits.max();
	const long minlong = long_limits.min();
	const long maxlong = long_limits.max();
#else
	const short minshort = std::numeric_limits<short>::min();
	const short maxshort = std::numeric_limits<short>::max();
	const long minlong = std::numeric_limits<long>::min();
	const long maxlong = std::numeric_limits<long>::max();
#endif

};

void XSDA::Free(void)
{
	if (mDescrArea != 0)
	{
		for (int i = 0; i < mDescrArea->sqln; i++)
		{
			XSQLVAR* var = &(mDescrArea->sqlvar[i]);
			if (var->sqldata != 0)
			{
				switch (var->sqltype & ~1)
				{
					case SQL_ARRAY :
					case SQL_BLOB :		delete (ISC_QUAD*) var->sqldata; break;
					case SQL_TIMESTAMP :delete (ISC_TIMESTAMP*) var->sqldata; break;
					case SQL_TYPE_TIME :delete (ISC_TIME*) var->sqldata; break;
					case SQL_TYPE_DATE :delete (ISC_DATE*) var->sqldata; break;
					case SQL_TEXT :
					case SQL_VARYING :	delete [] var->sqldata; break;
					case SQL_SHORT :	delete (short*) var->sqldata; break;
					case SQL_LONG :		delete (long*) var->sqldata; break;
					case SQL_INT64 :	delete (__int64*) var->sqldata; break;
					case SQL_FLOAT : 	delete (float*) var->sqldata; break;
					case SQL_DOUBLE :	delete (double*) var->sqldata; break;
					default : throw ExceptionImpl("XSDA::Free",
								"Found an unknown sqltype !");
				}
			}
			if (var->sqlind != 0) delete var->sqlind;
		}
		delete [] (char*)mDescrArea;
		mDescrArea = 0;
	}

	mNumerics.clear();
	mFloats.clear();
	mLarges.clear();
	mLongs.clear();
	mShorts.clear();
	mBools.clear();
	mStrings.clear();
}

void XSDA::Resize(short n)
{
	int size = XSQLDA_LENGTH(n);

	Free();
    mDescrArea = (XSQLDA*) new char[size];

	memset(mDescrArea, 0, size);
	mNumerics.resize(n);
	mFloats.resize(n);
	mLarges.resize(n);
	mLongs.resize(n);
	mShorts.resize(n);
	mBools.resize(n);
	mStrings.resize(n);
	for (int i = 0; i < n; i++)
	{
		mNumerics[i] = 0.0;
		mFloats[i] = 0.0;
		mLarges[i] = 0;
		mLongs[i] = 0;
		mShorts[i] = 0;
		mBools[i] = 0;
		mStrings[i].erase();
	}

	mDescrArea->version = SQLDA_VERSION1;
	mDescrArea->sqln = n;
}

void XSDA::AllocVariables(void)
{
	int i;
	for (i = 0; i < mDescrArea->sqld; i++)
	{
		XSQLVAR* var = &(mDescrArea->sqlvar[i]);
		switch (var->sqltype & ~1)
		{
			case SQL_ARRAY :
			case SQL_BLOB :		var->sqldata = (char*) new ISC_QUAD;
								memset(var->sqldata, 0, sizeof(ISC_QUAD));
								break;
			case SQL_TIMESTAMP :var->sqldata = (char*) new ISC_TIMESTAMP;
								memset(var->sqldata, 0, sizeof(ISC_TIMESTAMP));
								break;
			case SQL_TYPE_TIME :var->sqldata = (char*) new ISC_TIME;
								memset(var->sqldata, 0, sizeof(ISC_TIME));
								break;
			case SQL_TYPE_DATE :var->sqldata = (char*) new ISC_DATE;
								memset(var->sqldata, 0, sizeof(ISC_DATE));
								break;
			case SQL_TEXT :		var->sqldata = new char[var->sqllen+1];
								memset(var->sqldata, ' ', var->sqllen);
								var->sqldata[var->sqllen] = '\0';
								break;
			case SQL_VARYING :	var->sqldata = new char[var->sqllen+3];
								memset(var->sqldata, 0, 2);
								memset(var->sqldata+2, ' ', var->sqllen);
								var->sqldata[var->sqllen+2] = '\0';
								break;
			case SQL_SHORT :	var->sqldata = (char*) new short(0); break;
			case SQL_LONG :		var->sqldata = (char*) new long(0); break;
			case SQL_INT64 :	var->sqldata = (char*) new __int64(0); break;
			case SQL_FLOAT : 	var->sqldata = (char*) new float(0.0); break;
			case SQL_DOUBLE :	var->sqldata = (char*) new double(0.0); break;
			default : throw ExceptionImpl("XSDA::AllocVariables",
						"Found an unknown sqltype !");
		}
		if (var->sqltype & 1) var->sqlind = new short(-1);	// 0 indicator
	}
}

void XSDA::SetNull(int varnum)
{
	if (varnum < 1 || varnum > mDescrArea->sqld)
		throw ExceptionImpl("XSDA::SetNull", "Variable index out of range.");

	XSQLVAR* var = &(mDescrArea->sqlvar[varnum-1]);
	if (! (var->sqltype & 1))
		throw ExceptionImpl("XSDA::SetNull", "This parameter can't be null.");

	*var->sqlind = -1;	// Set the parameter to SQL NULL
}

void XSDA::SetValue(int varnum, IITYPE ivType, const void* value, int userlen)
{
	if (varnum < 1 || varnum > mDescrArea->sqld)
		throw ExceptionImpl("XSDA::SetValue", "Variable index out of range.");
	if (value == 0)
		throw ExceptionImpl("XSDA::SetValue", "Unexpected null pointer detected.");

	short len;
	XSQLVAR* var = &(mDescrArea->sqlvar[varnum-1]);
	switch (var->sqltype & ~1)
	{
		case SQL_TEXT :
			if (ivType == ivString)
			{
				std::string* svalue = (std::string*)value;
				len = (short)svalue->length();
				if (len > var->sqllen) len = var->sqllen;
				strncpy(var->sqldata, svalue->c_str(), len);
				while (len < var->sqllen) var->sqldata[len++] = ' ';
			}
			else if (ivType == ivByte)
			{
				if (userlen > var->sqllen) userlen = var->sqllen;
				memcpy(var->sqldata, value, userlen);
				while (userlen < var->sqllen) var->sqldata[userlen++] = ' ';
			}
			else if (ivType == ivDBKey)
			{
				IBPP::DBKey* key = (IBPP::DBKey*)value;
				key->GetKey(var->sqldata, var->sqllen);
			}
			else if (ivType == ivBool)
			{
				var->sqldata[0] = *(bool*)value ? 'T' : 'F';
				len = 1;
				while (len < var->sqllen) var->sqldata[len++] = ' ';
			}
			else throw ExceptionImpl("XSDA::SetValue", "Incompatible types.");
			break;

		case SQL_VARYING :
			if (ivType == ivString)
			{
				std::string* svalue = (std::string*)value;
				len = (short)svalue->length();
				if (len > var->sqllen) len = var->sqllen;
				*(short*)var->sqldata = (short)len;
				strncpy(var->sqldata+2, svalue->c_str(), len);
			}
			else if (ivType == ivByte)
			{
				if (userlen > var->sqllen) userlen = var->sqllen;
				*(short*)var->sqldata = (short)userlen;
				memcpy(var->sqldata+2, value, userlen);
			}
			else if (ivType == ivBool)
			{
				*(short*)var->sqldata = (short)1;
				var->sqldata[2] = *(bool*)value ? 'T' : 'F';
			}
			else throw ExceptionImpl("XSDA::SetValue", "Incompatible types.");
			break;

		case SQL_SHORT :
			if (ivType == ivBool)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(short*)var->sqldata = short(*(bool*)value ? 1 : 0);
			}
			else if (ivType == ivShort)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(short*)var->sqldata = *(short*)value;
			}
			else if (ivType == ivLong)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				if (*(long*)value < minshort || *(long*)value > maxshort)
					throw ExceptionImpl("XSDA::SetValue",
						"Out of range numeric conversion !");
				*(short*)var->sqldata = (short)*(long*)value;
			}
			else if (ivType == ivLarge)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				if (*(__int64*)value < minshort || *(__int64*)value > maxshort)
					throw ExceptionImpl("XSDA::SetValue",
						"Out of range numeric conversion !");
				*(short*)var->sqldata = (short)*(__int64*)value;
			}
			else if (ivType == ivFloat)
			{
				// This SQL_SHORT is a NUMERIC(x,y), scale it !
				double multiplier = dscales[-var->sqlscale];
				*(short*)var->sqldata =
					(short)floor(*(float*)value * multiplier + 0.5);
			}
			else if (ivType == ivDouble)
			{
				// This SQL_SHORT is a NUMERIC(x,y), scale it !
				double multiplier = dscales[-var->sqlscale];
				*(short*)var->sqldata =
					(short)floor(*(double*)value * multiplier + 0.5);
			}
			else throw ExceptionImpl("XSDA::SetValue", "Incompatible types.");
			break;

		case SQL_LONG :
			if (ivType == ivBool)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(long*)var->sqldata = *(bool*)value ? 1L : 0L;
			}
			else if (ivType == ivLong)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(long*)var->sqldata = *(long*)value;
			}
			else if (ivType == ivShort)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(long*)var->sqldata = *(short*)value;
			}
			else if (ivType == ivLarge)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				if (*(__int64*)value < minlong || *(__int64*)value > maxlong)
					throw ExceptionImpl("XSDA::SetValue",
						"Out of range numeric conversion !");
				*(long*)var->sqldata = (long)*(__int64*)value;
			}
			else if (ivType == ivFloat)
			{
				// This SQL_LONG is a NUMERIC(x,y), scale it !
				double multiplier = dscales[-var->sqlscale];
				*(long*)var->sqldata =
					(long)floor(*(float*)value * multiplier + 0.5);
			}
			else if (ivType == ivDouble)
			{
				// This SQL_LONG is a NUMERIC(x,y), scale it !
				double multiplier = dscales[-var->sqlscale];
				*(long*)var->sqldata =
					(long)floor(*(double*)value * multiplier + 0.5);
			}
			else throw ExceptionImpl("XSDA::SetValue", "Incompatible types.");
			break;

		case SQL_INT64 :
			if (ivType == ivBool)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::SetValue",
						"Numeric(x,y) field : use IStatement::GetDouble()");
				*(__int64*)var->sqldata = *(bool*)value ? 1 : 0;
			}
			else if (ivType == ivLarge)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::SetValue",
						"Numeric(x,y) field : use IStatement::GetDouble()");
				*(__int64*)var->sqldata = *(__int64*)value;
			}
			else if (ivType == ivShort)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(__int64*)var->sqldata = *(short*)value;
			}
			else if (ivType == ivLong)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::SetValue",
						"NUM/DEC with scale : use SetDouble()");
				*(__int64*)var->sqldata = *(long*)value;
			}
			else if (ivType == ivFloat)
			{
				// This SQL_INT64 is a NUMERIC(x,y), scale it !
				double multiplier = dscales[-var->sqlscale];
				*(__int64*)var->sqldata =
					(__int64)floor(*(float*)value * multiplier + 0.5);
			}
			else if (ivType == ivDouble)
			{
				// This SQL_INT64 is a NUMERIC(x,y), scale it !
				double multiplier = dscales[-var->sqlscale];
				*(__int64*)var->sqldata =
					(__int64)floor(*(double*)value * multiplier + 0.5);
			}
			else throw ExceptionImpl("XSDA::SetValue", "Incompatible types.");
			break;

		case SQL_FLOAT :
			if (ivType != ivFloat || var->sqlscale != 0)
				throw ExceptionImpl("XSDA::SetValue", "Incompatible types.");
			*(float*)var->sqldata = *(float*)value;
			break;

		case SQL_DOUBLE :
			if (ivType != ivDouble) throw ExceptionImpl("XSDA::SetValue",
										"Incompatible types.");
			if (var->sqlscale != 0)
			{
				// Round to scale of NUMERIC(x,y)
				double multiplier = dscales[-var->sqlscale];
				*(double*)var->sqldata =
					floor(*(double*)value * multiplier + 0.5) / multiplier;
			}
			else *(double*)var->sqldata = *(double*)value;
			break;

		case SQL_TIMESTAMP :
			if (ivType != ivTimestamp)
				throw ExceptionImpl("XSDA::SetValue", "Incompatible types.");
			{
				IBPP::Timestamp* timestamp = (IBPP::Timestamp*)value;
				//timestamp->GetNative((ISC_TIMESTAMP*)var->sqldata);
				int year, month, day;
				int hour, minute, second;
				timestamp->GetDate(year, month, day);
				timestamp->GetTime(hour, minute, second);
				tm tms;
				memset(&tms, 0, sizeof(tms));
				tms.tm_year = year - 1900;
				tms.tm_mon = month - 1;
				tms.tm_mday = day;
				tms.tm_hour = hour;
				tms.tm_min = minute;
				tms.tm_sec = second;
				(*gds.Call()->m_encode_timestamp)(&tms, (ISC_TIMESTAMP*)var->sqldata);
			}
			break;

		case SQL_TYPE_DATE :
			if (ivType != ivDate) throw ExceptionImpl("XSDA::SetValue",
										"Incompatible types.");
			{
				IBPP::Date* date = (IBPP::Date*)value;
				//date->GetNative((ISC_DATE*)var->sqldata);
				int year, month, day;
				date->GetDate(year, month, day);
				tm tms;
				memset(&tms, 0, sizeof(tms));
				tms.tm_year = year - 1900;
				tms.tm_mon = month - 1;
				tms.tm_mday = day;
				(*gds.Call()->m_encode_sql_date)(&tms, (ISC_DATE*)var->sqldata);
			}
			break;

		case SQL_TYPE_TIME :
			if (ivType != ivTime) throw ExceptionImpl("XSDA::SetValue",
										"Incompatible types.");
			{
				IBPP::Time* time = (IBPP::Time*)value;
				//time->GetNative((ISC_TIME*)var->sqldata);
				int hour, minute, second;
				time->GetTime(hour, minute, second);
				tm tms;
				memset(&tms, 0, sizeof(tms));
				tms.tm_hour = hour;
				tms.tm_min = minute;
				tms.tm_sec = second;
				(*gds.Call()->m_encode_sql_time)(&tms, (ISC_TIME*)var->sqldata);
			}
			break;

		case SQL_BLOB :
			if (ivType != ivBlob) throw ExceptionImpl("XSDA::SetValue",
										"Incompatible types.");
			{
				BlobImpl* blob = (BlobImpl*)value;
				blob->GetId((ISC_QUAD*)var->sqldata);
			}
			break;

		case SQL_ARRAY :
			if (ivType != ivArray) throw ExceptionImpl("XSDA::SetValue",
										"Incompatible types.");
			{
				ArrayImpl* array = (ArrayImpl*)value;
				array->GetId((ISC_QUAD*)var->sqldata);
				// When an array has been affected to a column, we want to reset
				// its ID. This way, the next WriteFrom() on the same Array object
				// will allocate a new ID. This protects against storing the same
				// array ID in multiple columns or rows.
				array->ResetId();
			}
			break;

		default : throw ExceptionImpl("XSDA::SetValue",
						"The field uses an unsupported SQL type !");
	}

	if (var->sqltype & 1) *var->sqlind = 0;		// Remove the 0 flag
}

bool XSDA::IsNull(int varnum)
{
	if (varnum < 1 || varnum > mDescrArea->sqld)
		throw ExceptionImpl("XSDA::IsNull", "Variable index out of range.");

	XSQLVAR* var = &(mDescrArea->sqlvar[varnum-1]);
	return ((var->sqltype & 1) && *(var->sqlind) != 0) ? true : false;
}

void* XSDA::GetValue(int varnum, IITYPE ivType, void* retvalue)
{
	if (varnum < 1 || varnum > mDescrArea->sqld)
		throw ExceptionImpl("XSDA::GetValue", "Variable index out of range.");

	void* value;
	int len;
	XSQLVAR* var = &(mDescrArea->sqlvar[varnum-1]);

	// When there is no value (SQL NULL)
	if ((var->sqltype & 1) && *(var->sqlind) != 0) return 0;

	switch (var->sqltype & ~1)
	{
		case SQL_TEXT :
			if (ivType == ivString)
			{
				mStrings[varnum-1].erase();
				mStrings[varnum-1].append(var->sqldata, var->sqllen);
				value = &mStrings[varnum-1];
			}
			else if (ivType == ivByte)
			{
				// In case of ivByte, void* retvalue points to an int where we
				// will store the len of the available data
				if (retvalue != 0) *(int*)retvalue = var->sqllen;
				value = var->sqldata;
			}
			else if (ivType == ivDBKey)
			{
				IBPP::DBKey* key = (IBPP::DBKey*)retvalue;
				key->SetKey(var->sqldata, var->sqllen);
				value = retvalue;
			}
			else if (ivType == ivBool)
			{
				mBools[varnum-1] = 0;
				if (var->sqllen >= 1)
				{
					char c = var->sqldata[0];
					if (c == 't' || c == 'T' || c == 'y' || c == 'Y' ||	c == '1')
						mBools[varnum-1] = 1;
				}
				value = &mBools[varnum-1];
			}
			else throw ExceptionImpl("XSDA::GetValue", "Incompatible types.");
			break;

		case SQL_VARYING :
			if (ivType == ivString)
			{
				mStrings[varnum-1].erase();
				mStrings[varnum-1].append(var->sqldata+2, (int)*(short*)var->sqldata);
				value = &mStrings[varnum-1];
			}
			else if (ivType == ivByte)
			{
				// In case of ivByte, void* retvalue points to an int where we
				// will store the len of the available data
				if (retvalue != 0) *(int*)retvalue = (int)*(short*)var->sqldata;
				value = var->sqldata+2;
			}
			else if (ivType == ivBool)
			{
				mBools[varnum-1] = 0;
				len = *(short*)var->sqldata;
				if (len >= 1)
				{
					char c = var->sqldata[2];
					if (c == 't' || c == 'T' || c == 'y' || c == 'Y' ||	c == '1')
						mBools[varnum-1] = 1;
				}
				value = &mBools[varnum-1];
			}
			else throw ExceptionImpl("XSDA::GetValue",
										"Incompatible types.");
			break;

		case SQL_SHORT :
			if (ivType == ivShort)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				value = var->sqldata;
			}
			else if (ivType == ivBool)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				if (*(short*)var->sqldata == 0)
					mBools[varnum-1] = 0;
				else mBools[varnum-1] = 1;
				value = &mBools[varnum-1];
			}
			else if (ivType == ivLong)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				mLongs[varnum-1] = *(short*)var->sqldata;
				value = &mLongs[varnum-1];
			}
			else if (ivType == ivLarge)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				mLarges[varnum-1] = *(short*)var->sqldata;
				value = &mLarges[varnum-1];
			}
			else if (ivType == ivFloat)
			{
				// This SQL_SHORT is a NUMERIC(x,y), scale it !
				double divisor = dscales[-var->sqlscale];
				mFloats[varnum-1] = (float)(*(short*)var->sqldata / divisor);

				value = &mFloats[varnum-1];
			}
			else if (ivType == ivDouble)
			{
				// This SQL_SHORT is a NUMERIC(x,y), scale it !
				double divisor = dscales[-var->sqlscale];
				mNumerics[varnum-1] = *(short*)var->sqldata / divisor;
				value = &mNumerics[varnum-1];
			}
			else throw ExceptionImpl("XSDA::GetValue", "Incompatible types.");
			break;

		case SQL_LONG :
			if (ivType == ivLong)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				value = var->sqldata;
			}
			else if (ivType == ivBool)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				if (*(long*)var->sqldata == 0L)
					mBools[varnum-1] = 0;
				else mBools[varnum-1] = 1;
				value = &mBools[varnum-1];
			}
			else if (ivType == ivShort)
			{
				long tmp;
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				tmp = *(long*)var->sqldata;
				if (tmp < minshort || tmp > maxshort)
					throw ExceptionImpl("XSDA::GetValue",
						"Out of range numeric conversion !");
				mShorts[varnum-1] = (short)tmp;
				value = &mShorts[varnum-1];
			}
			else if (ivType == ivLarge)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				mLarges[varnum-1] = *(long*)var->sqldata;
				value = &mLarges[varnum-1];
			}
			else if (ivType == ivFloat)
			{
				// This SQL_LONG is a NUMERIC(x,y), scale it !
				double divisor = dscales[-var->sqlscale];
				mFloats[varnum-1] = (float)(*(long*)var->sqldata / divisor);
				value = &mFloats[varnum-1];
			}
			else if (ivType == ivDouble)
			{
				// This SQL_LONG is a NUMERIC(x,y), scale it !
				double divisor = dscales[-var->sqlscale];
				mNumerics[varnum-1] = *(long*)var->sqldata / divisor;
				value = &mNumerics[varnum-1];
			}
			else throw ExceptionImpl("XSDA::GetValue", "Incompatible types.");
			break;

		case SQL_INT64 :
			if (ivType == ivLarge)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				value = var->sqldata;
			}
			else if (ivType == ivBool)
			{
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				if (*(__int64*)var->sqldata == 0)
					mBools[varnum-1] = 0;
				else mBools[varnum-1] = 1;
				value = &mBools[varnum-1];
			}
			else if (ivType == ivShort)
			{
				__int64 tmp;
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				tmp = *(__int64*)var->sqldata;
				if (tmp < minshort || tmp > maxshort)
					throw ExceptionImpl("XSDA::GetValue",
						"Out of range numeric conversion !");
				mShorts[varnum-1] = (short)tmp;
				value = &mShorts[varnum-1];
			}
			else if (ivType == ivLong)
			{
				__int64 tmp;
				if (var->sqlscale != 0)
					throw ExceptionImpl("XSDA::GetValue",
						"NUM/DEC with scale : use GetDouble()");
				tmp = *(__int64*)var->sqldata;
				if (tmp < minlong || tmp > maxlong)
					throw ExceptionImpl("XSDA::GetValue",
						"Out of range numeric conversion !");
				mLongs[varnum-1] = (long)tmp;
				value = &mLongs[varnum-1];
			}
			else if (ivType == ivFloat)
			{
				// This SQL_INT64 is a NUMERIC(x,y), scale it !
				double divisor = dscales[-var->sqlscale];
				mFloats[varnum-1] = (float)(*(__int64*)var->sqldata / divisor);
				value = &mFloats[varnum-1];
			}
			else if (ivType == ivDouble)
			{
				// This SQL_INT64 is a NUMERIC(x,y), scale it !
				double divisor = dscales[-var->sqlscale];
				mNumerics[varnum-1] = *(__int64*)var->sqldata / divisor;
				value = &mNumerics[varnum-1];
			}
			else throw ExceptionImpl("XSDA::GetValue", "Incompatible types.");
			break;

		case SQL_FLOAT :
			if (ivType != ivFloat) throw ExceptionImpl("XSDA::GetValue",
										"Incompatible types.");
			value = var->sqldata;
			break;

		case SQL_DOUBLE :
			if (ivType != ivDouble) throw ExceptionImpl("XSDA::GetValue",
										"Incompatible types.");
			if (var->sqlscale != 0)
			{
				// Round to scale y of NUMERIC(x,y)
				double multiplier = dscales[-var->sqlscale];
				mNumerics[varnum-1] =
					floor(*(double*)var->sqldata * multiplier + 0.5) / multiplier;
				value = &mNumerics[varnum-1];
			}
			else value = var->sqldata;
			break;

		case SQL_TIMESTAMP :
			if (ivType != ivTimestamp)
				throw ExceptionImpl("XSDA::SetValue", "Incompatible types.");
			{
				IBPP::Timestamp* timestamp = (IBPP::Timestamp*)retvalue;
				//timestamp->SetNative((ISC_TIMESTAMP*)var->sqldata);
				tm tms;
				(*gds.Call()->m_decode_timestamp)((ISC_TIMESTAMP*)var->sqldata, &tms);
				timestamp->SetDate(tms.tm_year + 1900, tms.tm_mon + 1, tms.tm_mday);
				timestamp->SetTime(tms.tm_hour, tms.tm_min, tms.tm_sec);
				value = retvalue;
			}
			break;

		case SQL_TYPE_DATE :
			if (ivType != ivDate) throw ExceptionImpl("XSDA::SetValue",
										"Incompatible types.");
			{
				IBPP::Date* date = (IBPP::Date*)retvalue;
				//date->SetNative((ISC_DATE*)var->sqldata);
				tm tms;
				(*gds.Call()->m_decode_sql_date)((ISC_DATE*)var->sqldata, &tms);
				date->SetDate(tms.tm_year + 1900, tms.tm_mon + 1, tms.tm_mday);
				value = retvalue;
			}
			break;

		case SQL_TYPE_TIME :
			if (ivType != ivTime) throw ExceptionImpl("XSDA::SetValue",
										"Incompatible types.");
			{
				IBPP::Time* time = (IBPP::Time*)retvalue;
				//time->SetNative((ISC_TIME*)var->sqldata);
				tm tms;
				(*gds.Call()->m_decode_sql_time)((ISC_TIME*)var->sqldata, &tms);
				time->SetTime(tms.tm_hour, tms.tm_min, tms.tm_sec);
				value = retvalue;
			}
			break;

		case SQL_BLOB :
			if (ivType != ivBlob) throw ExceptionImpl("XSDA::GetValue",
										"Incompatible types.");
			{
				BlobImpl* blob = (BlobImpl*)retvalue;
				blob->SetId((ISC_QUAD*)var->sqldata);
				value = retvalue;
			}
			break;

		case SQL_ARRAY :
			if (ivType != ivArray) throw ExceptionImpl("XSDA::GetValue",
										"Incompatible types.");
			{
				ArrayImpl* array = (ArrayImpl*)retvalue;
				array->SetId((ISC_QUAD*)var->sqldata);
				value = retvalue;
			}
			break;

		default : throw ExceptionImpl("XSDA::GetValue",
						"Found an unknown sqltype !");
	}

	return value;
}

int XSDA::GetVarNum(const std::string& name)
{
	XSQLVAR* var;
	char Uname[sizeof(var->sqlname)+1];		// Max size of sqlname + '\0'

	if (name.empty())
		throw ExceptionImpl("XSDA::GetVarNum", "Unexpected null pointer detected.");

	// Local upper case copy of the column name
	size_t len = name.length();
	if (len > sizeof(var->sqlname)) len = sizeof(var->sqlname);
	strncpy(Uname, name.c_str(), len);
	Uname[len] = '\0';
	char* p = Uname;
	while (*p != '\0') { *p = char(toupper(*p)); ++p; }

	// Loop through the columns of the descriptor
	for (int i = 0; i < mDescrArea->sqld; i++)
	{
		var = &(mDescrArea->sqlvar[i]);
		if (var->sqlname_length != (short)len) continue;
		if (strncmp(Uname, var->sqlname, len) == 0) return i+1;
	}

	return GetVarNumAlias(name);
}

int XSDA::GetVarNumAlias(const std::string& alias)
{
	XSQLVAR* var;
	char Ualias[sizeof(var->aliasname)+1];		// Max size of aliasname + '\0'

	if (alias.empty())
		throw ExceptionImpl("XSDA::GetVarNumAlias", "Unexpected null pointer detected.");

	// Local upper case copy of the column name
	size_t len = alias.length();
	if (len > sizeof(var->aliasname)) len = sizeof(var->aliasname);
	strncpy(Ualias, alias.c_str(), len);
	Ualias[len] = '\0';
	char* p = Ualias;
	while (*p != '\0') { *p = char(toupper(*p)); ++p; }

	// Loop through the columns of the descriptor
	for (int i = 0; i < mDescrArea->sqld; i++)
	{
		var = &(mDescrArea->sqlvar[i]);
		if (var->aliasname_length != (short)len) continue;
		if (strncmp(Ualias, var->aliasname, len) == 0) return i+1;
	}

	return 0;
}
/*
GetName, GetAlias, GetTable : all these 3 have a mistake.
Ideally, the strings should be stored elsewhere (like _Numerics and so on) to
take into account the final '\0' which needs to be added. For now, we insert
the '\0' in the original data, which will cut the 32th character. Not terribly
bad, but should be cleanly rewritten.
*/

const char* XSDA::GetName(int varnum)
{
	if (varnum < 1 || varnum > mDescrArea->sqld)
		throw ExceptionImpl("XSDA::GetName", "Variable index out of range.");

	XSQLVAR* var = &(mDescrArea->sqlvar[varnum-1]);
	if (var->sqlname_length >= 31) var->sqlname_length = 31;
	var->sqlname[var->sqlname_length] = '\0';
	return var->sqlname;
}

const char* XSDA::GetAlias(int varnum)
{
	if (varnum < 1 || varnum > mDescrArea->sqld)
		throw ExceptionImpl("XSDA::GetAlias", "Variable index out of range.");

	XSQLVAR* var = &(mDescrArea->sqlvar[varnum-1]);
	if (var->aliasname_length >= 31) var->aliasname_length = 31;
	var->aliasname[var->aliasname_length] = '\0';
	return var->aliasname;
}

const char* XSDA::GetTable(int varnum)
{
	if (varnum < 1 || varnum > mDescrArea->sqld)
		throw ExceptionImpl("XSDA::GetTable", "Variable index out of range.");

	XSQLVAR* var = &(mDescrArea->sqlvar[varnum-1]);
	if (var->relname_length >= 31) var->relname_length = 31;
	var->relname[var->relname_length] = '\0';
	return var->relname;
}

IBPP::SDT XSDA::GetType(int varnum)
{
	if (varnum < 1 || varnum > mDescrArea->sqld)
		throw ExceptionImpl("XSDA::GetType", "Variable index out of range.");

	IBPP::SDT value;
	XSQLVAR* var = &(mDescrArea->sqlvar[varnum-1]);

	switch (var->sqltype & ~1)
	{
		case SQL_TEXT :      value = IBPP::sdString;    break;
		case SQL_VARYING :   value = IBPP::sdString;    break;
		case SQL_SHORT :     value = IBPP::sdSmallint;  break;
		case SQL_LONG :      value = IBPP::sdInteger;   break;
		case SQL_INT64 :     value = IBPP::sdLargeint;  break;
		case SQL_FLOAT :     value = IBPP::sdFloat;     break;
		case SQL_DOUBLE :    value = IBPP::sdDouble;    break;
		case SQL_TIMESTAMP : value = IBPP::sdTimestamp; break;
		case SQL_TYPE_DATE : value = IBPP::sdDate;      break;
		case SQL_TYPE_TIME : value = IBPP::sdTime;      break;
		case SQL_BLOB :      value = IBPP::sdBlob;      break;
		case SQL_ARRAY :     value = IBPP::sdArray;     break;
		default : throw ExceptionImpl("XSDA::GetType",
						"Found an unknown sqltype !");
	}

	return value;
}

int XSDA::GetSize(int varnum)
{
	if (varnum < 1 || varnum > mDescrArea->sqld)
		throw ExceptionImpl("XSDA::GetSize", "Variable index out of range.");

	XSQLVAR* var = &(mDescrArea->sqlvar[varnum-1]);
    return var->sqllen;
}

int XSDA::GetScale(int varnum)
{
	if (varnum < 1 || varnum > mDescrArea->sqld)
		throw ExceptionImpl("XSDA::GetScale", "Variable index out of range.");

	XSQLVAR* var = &(mDescrArea->sqlvar[varnum-1]);
    return -var->sqlscale;
}

XSDA::XSDA()
	: mDescrArea(0)
{
	Resize(1);
}

XSDA::XSDA(const XSDA& copied)
{
	int n = copied.mDescrArea->sqln;
	int size = XSQLDA_LENGTH(n);
	int i;

	// Initial brute copy
    mDescrArea = (XSQLDA*) new char[size];
	memcpy(mDescrArea, copied.mDescrArea, size);

	// Pointers init, real data copy
	mNumerics = copied.mNumerics;
	mFloats = copied.mFloats;
	mLarges = copied.mLarges;
	mLongs = copied.mLongs;
	mShorts = copied.mShorts;
	mBools = copied.mBools;
	mStrings = copied.mStrings;

	// Copy of the columns data
	for (i = 0; i < mDescrArea->sqld; i++)
	{
		XSQLVAR* var = &(mDescrArea->sqlvar[i]);
		XSQLVAR* org = &(copied.mDescrArea->sqlvar[i]);
		switch (var->sqltype & ~1)
		{
			case SQL_ARRAY :
			case SQL_BLOB :		var->sqldata = (char*) new ISC_QUAD;
								memcpy(var->sqldata, org->sqldata, sizeof(ISC_QUAD));
								break;
			case SQL_TIMESTAMP :var->sqldata = (char*) new ISC_TIMESTAMP;
								memcpy(var->sqldata, org->sqldata, sizeof(ISC_TIMESTAMP));
								break;
			case SQL_TYPE_TIME :var->sqldata = (char*) new ISC_TIME;
								memcpy(var->sqldata, org->sqldata, sizeof(ISC_TIME));
								break;
			case SQL_TYPE_DATE :var->sqldata = (char*) new ISC_DATE;
								memcpy(var->sqldata, org->sqldata, sizeof(ISC_DATE));
								break;
			case SQL_TEXT :		var->sqldata = new char[var->sqllen+1];
								memcpy(var->sqldata, org->sqldata, var->sqllen+1);
								break;
			case SQL_VARYING :	var->sqldata = new char[var->sqllen+3];
								memcpy(var->sqldata, org->sqldata, var->sqllen+3);
								break;
			case SQL_SHORT :	var->sqldata = (char*) new short(*(short*)org->sqldata); break;
			case SQL_LONG :		var->sqldata = (char*) new long(*(long*)org->sqldata); break;
			case SQL_INT64 :	var->sqldata = (char*) new __int64(*(__int64*)org->sqldata); break;
			case SQL_FLOAT : 	var->sqldata = (char*) new float(*(float*)org->sqldata); break;
			case SQL_DOUBLE :	var->sqldata = (char*) new double(*(double*)org->sqldata); break;
			default : throw ExceptionImpl("XSDA::Ctor",
						"Found an unknown sqltype !");
		}
		if (var->sqltype & 1) var->sqlind = new short(*org->sqlind);	// 0 indicator
	}
}

XSDA::XSDA(short n)
	: mDescrArea(0)
{
	Resize(n);
}

XSDA::~XSDA()
{
	Free();
}

//
//	EOF
//
