///////////////////////////////////////////////////////////////////////////////
//
//	File    : $Id: statement.cpp,v 1.30.2.4 2004/09/23 15:34:08 epocman Exp $
//	Subject : IBPP, Service class implementation
//
///////////////////////////////////////////////////////////////////////////////
//
//	The contents of this file are subject to the Mozilla Public License
//	Version 1.0 (the "License"); you may not use this file except in
//	compliance with the License. You may obtain a copy of the License at
//	http://www.mozilla.org/MPL/
//
//	Software distributed under the License is distributed on an "AS IS"
//	basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
//	License for the specific language governing rights and limitations
//	under the License.
//
//	The Original Code is "IBPP 0.9" and all its associated documentation.
//
//	The Initial Developer of the Original Code is T.I.P. Group S.A.
//	Portions created by T.I.P. Group S.A. are
//	Copyright (C) 2000 T.I.P Group S.A.
//	All Rights Reserved.
//
//	Contributor(s): ______________________________________.
//
///////////////////////////////////////////////////////////////////////////////
//
//	COMMENTS
//	* Tabulations should be set every four characters when editing this file.
//
///////////////////////////////////////////////////////////////////////////////

#include "ibpp.h"
#include "_internals.h"

#ifdef HAS_HDRSTOP
#pragma hdrstop
#endif

using namespace ibpp_internals;

//	(((((((( OBJECT INTERFACE IMPLEMENTATION ))))))))

IBPP::IDatabase* StatementImpl::Database(void) const
{
	return mDatabase;
}

IBPP::ITransaction* StatementImpl::Transaction(void) const
{
	return mTransaction;
}

void StatementImpl::Prepare(const std::string& sql)
{
	if (mDatabase == 0)
		throw ExceptionImpl("Statement::Prepare", "An IDatabase must be attached.");
	if (mDatabase->GetHandle() == 0)
		throw ExceptionImpl("Statement::Prepare", "IDatabase must be connected.");
	if (mTransaction == 0)
		throw ExceptionImpl("Statement::Prepare", "An ITransaction must be attached.");
	if (mTransaction->GetHandle() == 0)
		throw ExceptionImpl("Statement::Prepare", "ITransaction must be started.");
	if (sql.empty())
		throw ExceptionImpl("Statement::Prepare", "SQL statement can't be 0.");

	// Saves the SQL sentence, only for reporting reasons in case of errors
	mSql = sql;
	
	IBS status;

	// Free all resources currently attached to this Statement, then allocate
	// a new statement descriptor.
	try {Close();} catch(IBPP::Exception&) {}
	(*gds.Call()->m_dsql_allocate_statement)(status.Self(), mDatabase->GetHandlePtr(), &mHandle);
	if (status.Errors())
		throw ExceptionImpl(status, "Statement::Prepare",
			"isc_dsql_allocate_statement failed");

	// Empirical estimate of parameters count and output columns count.
	// This is by far not an exact estimation, which would require parsing the
	// SQL statement. If the SQL statement contains '?' and ',' in string
	// constants, this count will obviously be wrong, but it will exagerated.
	// It won't hurt. We just try to not have to re-allocate those descriptors later.
	// So we prefer to get them a little bit larger than needed than the other way.
	short inEstimate = 0;
	short outEstimate = 1;
	for (int i = 0; i < (int)strlen(sql.c_str()); i++)
	{
		if (sql[i] == '?') ++inEstimate;
		if (sql[i] == ',') ++outEstimate;
	}

	DebugStream()<< "Prepare(\""<< sql<< "\")"<< fds;
	DebugStream()<< "Estimation: "<< inEstimate<< " IN parameters and "
			<< outEstimate<< " OUT columns"<< fds;

	// Allocates output descriptor and prepare the statement
	mOutDescr = new XSDA(outEstimate);
	status.Reset();
	(*gds.Call()->m_dsql_prepare)(status.Self(), mTransaction->GetHandlePtr(),
		&mHandle, (short)sql.length(), const_cast<char*>(sql.c_str()),
			short(mDatabase->Dialect()), mOutDescr->Self());
	if (status.Errors())
	{
		try {Close();} catch(IBPP::Exception&) {}
		std::string context = "Statement::Prepare( ";
		context.append(mSql).append(" )");
		throw ExceptionImpl(status, context.c_str(),
			"isc_dsql_prepare failed");
	}

	// Read what kind of statement was prepared
	status.Reset();
	char itemsReq[] = {isc_info_sql_stmt_type};
	char itemsRes[8];
	(*gds.Call()->m_dsql_sql_info)(status.Self(), &mHandle, 1, itemsReq,
		sizeof(itemsRes), itemsRes);
	if (status.Errors())
	{
		try {Close();} catch(IBPP::Exception&) {}
		throw ExceptionImpl(status, "Statement::Prepare",
			"isc_dsql_sql_info failed");
	}
	if (itemsRes[0] == isc_info_sql_stmt_type)
	{
		switch (itemsRes[3])
		{
			case isc_info_sql_stmt_select :		mType = IBPP::stSelect; break;
			case isc_info_sql_stmt_insert :		mType = IBPP::stInsert; break;
			case isc_info_sql_stmt_update :		mType = IBPP::stUpdate; break;
			case isc_info_sql_stmt_delete :		mType = IBPP::stDelete; break;
			case isc_info_sql_stmt_ddl :		mType = IBPP::stDDL; break;
			case isc_info_sql_stmt_exec_procedure : mType = IBPP::stExecProcedure; break;
			case isc_info_sql_stmt_select_for_upd : mType = IBPP::stSelectUpdate; break;
			case isc_info_sql_stmt_set_generator :	mType = IBPP::stOther; break;
			default : mType = IBPP::stUnsupported;
		}
	}
	if (mType == IBPP::stUnknown || mType == IBPP::stUnsupported)
	{
		try {Close();} catch(IBPP::Exception&) {}
		throw ExceptionImpl("Statement::Prepare",
			"Unknown or unsupported statement type");
	}

	if (mOutDescr->DescrSize() == 0)
	{
		// Get rid of the output descriptor, if it wasn't required (no output)
		delete mOutDescr;
		mOutDescr = 0;

		DebugStream()<< "Dropped output descriptor which was not required"<< fds;
	}
	else if (mOutDescr->DescrSize() > mOutDescr->AllocatedSize())
	{
		// Resize the output descriptor (which is too small).
		// The statement does not need to be prepared again, though the
		// output columns must be described again.

		DebugStream()<< "Resize output descriptor from "
			<< mOutDescr->AllocatedSize()<< " to "<< mOutDescr->DescrSize()<< fds;

		mOutDescr->Resize(mOutDescr->DescrSize());
		status.Reset();
		(*gds.Call()->m_dsql_describe)(status.Self(), &mHandle, 1, mOutDescr->Self());
		if (status.Errors())
		{
			try {Close();} catch(IBPP::Exception&) {}
			throw ExceptionImpl(status, "Statement::Prepare",
				"isc_dsql_describe failed");
		}
	}

	if (inEstimate > 0)
	{
		// Ready an input descriptor
		mInDescr = new XSDA(inEstimate);
		status.Reset();
		(*gds.Call()->m_dsql_describe_bind)(status.Self(), &mHandle, 1, mInDescr->Self());
		if (status.Errors())
		{
			try {Close();} catch(IBPP::Exception&) {}
			throw ExceptionImpl(status, "Statement::Prepare",
				"isc_dsql_describe_bind failed");
		}

		if (mInDescr->DescrSize() == 0)
		{
			// Get rid of the input descriptor, if it wasn't required (no parameters)
			delete mInDescr;
			mInDescr = 0;

			DebugStream()<< "Dropped input descriptor which was not required"<< fds;
		}
		else if (mInDescr->DescrSize() > mInDescr->AllocatedSize())
		{
			// Resize the input descriptor (which is too small).
			// The statement does not need to be prepared again, though the
			// parameters must be described again.

			DebugStream()<< "Resize input descriptor from "
					<< mInDescr->AllocatedSize()<< " to "
					<< mInDescr->DescrSize()<< fds;
			
			mInDescr->Resize(mInDescr->DescrSize());
			status.Reset();
			(*gds.Call()->m_dsql_describe_bind)(status.Self(), &mHandle, 1, mInDescr->Self());
			if (status.Errors())
			{
				try {Close();} catch(IBPP::Exception&) {}
				throw ExceptionImpl(status, "Statement::Prepare",
					"isc_dsql_describe_bind failed");
			}
		}
	}

	// Allocates variables of the input descriptor
	// 'InMissing' will later allow us to detect missing values for parameters
	if (mInDescr != 0)
	{
		mInDescr->AllocVariables();
		mInMissing = new bool[mInDescr->DescrSize()];
		for (int i = 0; i < mInDescr->DescrSize(); i++) mInMissing[i] = true;
	}

	// Allocates variables of the output descriptor
	if (mOutDescr != 0) mOutDescr->AllocVariables();
}

void StatementImpl::Plan(std::string& plan)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Plan", "No statement has been prepared.");
	if (mDatabase == 0)
		throw ExceptionImpl("Statement::Plan", "A Database must be attached.");
	if (mDatabase->GetHandle() == 0)
		throw ExceptionImpl("Statement::Plan", "Database must be connected.");

	IBS status;
	RB result(2048);
	char itemsReq[] = {isc_info_sql_get_plan};

	(*gds.Call()->m_dsql_sql_info)(status.Self(), &mHandle, 1, itemsReq,
								   result.Size(), result.Self());
	if (status.Errors()) throw ExceptionImpl(status,
								"Statement::Plan", "isc_dsql_sql_info failed.");

	result.GetString(isc_info_sql_get_plan, plan);
	if (plan[0] == '\n') plan.erase(0, 1);
}

void StatementImpl::Execute(const std::string& sql)
{
	if (! sql.empty()) Prepare(sql);

	if (mHandle == 0)
		throw ExceptionImpl("Statement::Execute", "No statement has been prepared.");

	// Check that a value has been set for each input parameter
	if (mInDescr != 0)
		for (int i = 0; i < mInDescr->DescrSize(); i++)
			if (mInMissing[i]) throw ExceptionImpl("Statement::Execute",
								"All parameters must be specified.");

	CursorFree();	// Free a previous 'cursor' if any

	IBS status;
	if (mType == IBPP::stSelect)
	{
		// Could return a result set (none, single or multi rows)
		(*gds.Call()->m_dsql_execute)(status.Self(), mTransaction->GetHandlePtr(),
			&mHandle, 1, mInDescr == 0 ? 0 : mInDescr->Self());
		if (status.Errors())
		{
			//Close();	Commented because Execute error should not free the statement
			std::string context = "Statement::Execute( ";
			context.append(mSql).append(" )");
			throw ExceptionImpl(status, context.c_str(),
				"isc_dsql_execute failed");
		}
		if (mOutDescr != 0) mResultSetAvailable = true;
	}
	else
	{
		// Should return at most a single row
		(*gds.Call()->m_dsql_execute2)(status.Self(), mTransaction->GetHandlePtr(),
			&mHandle, 1, mInDescr == 0 ? 0 : mInDescr->Self(),
			mOutDescr == 0 ? 0 : mOutDescr->Self());
		if (status.Errors())
		{
			//Close();	Commented because Execute error should not free the statement
			std::string context = "Statement::Execute( ";
			context.append(mSql).append(" )");
			throw ExceptionImpl(status, context.c_str(),
				"isc_dsql_execute2 failed");
		}
	}
}

void StatementImpl::CursorExecute(const std::string& cursor, const std::string& sql)
{
	if (cursor.empty())
		throw ExceptionImpl("Statement::CursorExecute", "Cursor name can't be 0.");

	if (! sql.empty()) Prepare(sql);

	if (mHandle == 0)
		throw ExceptionImpl("Statement::CursorExecute", "No statement has been prepared.");
	if (mType != IBPP::stSelectUpdate)
		throw ExceptionImpl("Statement::CursorExecute", "Statement must be a SELECT FOR UPDATE.");
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::CursorExecute", "Statement would return no rows.");

	// Check that a value has been set for each input parameter
	if (mInDescr != 0)
		for (int i = 0; i < mInDescr->DescrSize(); i++)
			if (mInMissing[i]) throw ExceptionImpl("Statement::CursorExecute",
								"All parameters must be specified.");

	CursorFree();	// Free a previous 'cursor' if any

	IBS status;
	(*gds.Call()->m_dsql_execute)(status.Self(), mTransaction->GetHandlePtr(),
		&mHandle, 1, mInDescr == 0 ? 0 : mInDescr->Self());
	if (status.Errors())
	{
		//Close();	Commented because Execute error should not free the statement
		std::string context = "Statement::CursorExecute( ";
		context.append(mSql).append(" )");
		throw ExceptionImpl(status, context.c_str(),
			"isc_dsql_execute failed");
	}

	status.Reset();
	(*gds.Call()->m_dsql_set_cursor_name)(status.Self(), &mHandle, const_cast<char*>(cursor.c_str()), 0);
	if (status.Errors())
	{
		//Close();	Commented because Execute error should not free the statement
		throw ExceptionImpl(status, "Statement::CursorExecute",
			"isc_dsql_set_cursor_name failed");
	}

	mResultSetAvailable = true;
}

void StatementImpl::ExecuteImmediate(const std::string& sql)
{
	if (mDatabase == 0)
		throw ExceptionImpl("Statement::ExecuteImmediate", "An IDatabase must be attached.");
	if (mDatabase->GetHandle() == 0)
		throw ExceptionImpl("Statement::ExecuteImmediate", "IDatabase must be connected.");
	if (mTransaction == 0)
		throw ExceptionImpl("Statement::ExecuteImmediate", "An ITransaction must be attached.");
	if (mTransaction->GetHandle() == 0)
		throw ExceptionImpl("Statement::ExecuteImmediate", "ITransaction must be started.");
	if (sql.empty())
		throw ExceptionImpl("Statement::ExecuteImmediate", "SQL statement can't be 0.");

	IBS status;
	try {Close();} catch(IBPP::Exception&) {}
    (*gds.Call()->m_dsql_execute_immediate)(status.Self(), mDatabase->GetHandlePtr(),
    	mTransaction->GetHandlePtr(), 0, const_cast<char*>(sql.c_str()),
    		short(mDatabase->Dialect()), 0);
    if (status.Errors())
	{
		try {Close();} catch(IBPP::Exception&) {}
		std::string context = "Statement::ExecuteImmediate( ";
		context.append(sql).append(" )");
		throw ExceptionImpl(status, context.c_str(),
			"isc_dsql_execute_immediate failed");
	}
}

int StatementImpl::AffectedRows(void)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::AffectedRows", "No statement has been prepared.");
	if (mDatabase == 0)
		throw ExceptionImpl("Statement::AffectedRows", "A Database must be attached.");
	if (mDatabase->GetHandle() == 0)
		throw ExceptionImpl("Statement::AffectedRows", "Database must be connected.");

	/*
	if (mType != IBPP::stInsert && mType != IBPP::stUpdate && mType != IBPP::stDelete)
	{
		// We return a count of 0 for any non interesting request. For selects, it is
		// assumed to be better to count the rows as they are fetched. So this IBPP API
		// can only be called after a successfull INSERT, UPDATE or DELETE.
		return 0;
	}
	*/

	int count;
	IBS status;
	RB result;
	char itemsReq[] = {isc_info_sql_records};

	(*gds.Call()->m_dsql_sql_info)(status.Self(), &mHandle, 1, itemsReq,
		result.Size(), result.Self());
	if (status.Errors()) throw ExceptionImpl(status,
			"Statement::AffectedRows", "isc_dsql_sql_info failed.");

	if (mType == IBPP::stInsert)
			count = result.GetValue(isc_info_sql_records, isc_info_req_insert_count);
	else if (mType == IBPP::stUpdate)
			count = result.GetValue(isc_info_sql_records, isc_info_req_update_count);
	else if (mType == IBPP::stDelete)
			count = result.GetValue(isc_info_sql_records, isc_info_req_delete_count);
	else if (mType == IBPP::stSelect)
			count = result.GetValue(isc_info_sql_records, isc_info_req_select_count);
	else	count = 0;	// Returns zero count for unknown cases

	return count;
}

bool StatementImpl::Fetch(void)
{
	if (! mResultSetAvailable)
		throw ExceptionImpl("Statement::Fetch",
			"No statement has been executed or no result set available.");

	IBS status;
	int code = (*gds.Call()->m_dsql_fetch)(status.Self(), &mHandle, 1, mOutDescr->Self());
	if (code == 100)	// This special code means "no more rows"
	{
		CursorFree();	// Free the explicit or implicit cursor/result-set
		return false;
	}
	if (status.Errors())
	{
		try {Close();} catch(IBPP::Exception&) {}
		throw ExceptionImpl(status, "Statement::Fetch",
			"isc_dsql_fetch failed.");
	}

	return true;
}

void StatementImpl::Close(void)
{
	// Free all statement resources. Used before preparing a new statement or
	// from destructor.
	if (mHandle != 0)
	{
		IBS status;
		(*gds.Call()->m_dsql_free_statement)(status.Self(), &mHandle, DSQL_drop);
		mHandle = 0;
		if (status.Errors())
			throw ExceptionImpl(status, "Statement::Close",
				"isc_dsql_free_statement failed.");
	}
	if (mInMissing != 0)
	{
		delete [] mInMissing;
		mInMissing = 0;
	}
	if (mInDescr != 0)
	{
		delete mInDescr;
		mInDescr = 0;
	}
	if (mOutDescr != 0)
	{
		delete mOutDescr;
		mOutDescr = 0;
	}

	mResultSetAvailable = false;
	mType = IBPP::stUnknown;
}

void StatementImpl::SetNull(int param)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::SetNull", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::SetNull", "The statement does not take parameters.");

	mInDescr->SetNull(param);
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, bool value)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[bool]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[bool]", "The statement does not take parameters.");

	mInDescr->SetValue(param, ivBool, &value);
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, const char* cstring)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[char*]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[char*]", "The statement does not take parameters.");
	if (cstring == 0)
		throw ExceptionImpl("Statement::Set[char*]", "0 char* pointer detected.");

	mInDescr->SetValue(param, ivByte, cstring, (int)strlen(cstring));
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, const void* bindata, int len)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[void*]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[void*]", "The statement does not take parameters.");
	if (bindata == 0)
		throw ExceptionImpl("Statement::Set[void*]", "0 char* pointer detected.");
	if (len < 0)
		throw ExceptionImpl("Statement::Set[void*]", "Length must be >= 0");
		
	mInDescr->SetValue(param, ivByte, bindata, len);
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, const std::string& s)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[string]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[string]", "The statement does not take parameters.");

	mInDescr->SetValue(param, ivString, (void*)&s);
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, short value)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[short]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[short]", "The statement does not take parameters.");
											
	mInDescr->SetValue(param, ivShort, &value);
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, int value)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[int]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[int]", "The statement does not take parameters.");

	mInDescr->SetValue(param, ivLong, &value);
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, long value)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[long]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[long]", "The statement does not take parameters.");

	mInDescr->SetValue(param, ivLong, &value);
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, __int64 value)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[int64]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[int64]", "The statement does not take parameters.");

	mInDescr->SetValue(param, ivLarge, &value);
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, float value)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[float]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[float]", "The statement does not take parameters.");

	mInDescr->SetValue(param, ivFloat, &value);
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, double value)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[double]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[double]", "The statement does not take parameters.");

	mInDescr->SetValue(param, ivDouble, &value);
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, const IBPP::Timestamp& value)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[Timestamp]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[Timestamp]", "The statement does not take parameters.");

	mInDescr->SetValue(param, ivTimestamp, &value);
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, const IBPP::Date& value)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[Date]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[Date]", "The statement does not take parameters.");

	if (mDatabase->Dialect() == 1)
	{
		// In dialect 1, IBPP::Date is supposed to work with old 'DATE'
		// fields which are actually ISC_TIMESTAMP.
		IBPP::Timestamp timestamp;
		timestamp = value;
		mInDescr->SetValue(param, ivTimestamp, &timestamp);
	}
	else
	{
		// Dialect 3
		mInDescr->SetValue(param, ivDate, (void*)&value);
	}

	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, const IBPP::Time& value)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[Time]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[Time]", "The statement does not take parameters.");
	if (mDatabase->Dialect() == 1)
		throw ExceptionImpl("Statement::Set[Time]", "Requires use of a dialect 3 database.");

	mInDescr->SetValue(param, ivTime, &value);
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, const IBPP::Blob& blob)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[Blob]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[Blob]", "The statement does not take parameters.");
	if (dynamic_cast<DatabaseImpl*>(blob->Database()) != mDatabase)
		throw ExceptionImpl("Statement::Set[Blob]",
			"IBlob and Statement attached to different databases");
	if (dynamic_cast<TransactionImpl*>(blob->Transaction()) != mTransaction)
		throw ExceptionImpl("Statement::Set[Blob]",
			"IBlob and Statement attached to different transactions");

	mInDescr->SetValue(param, ivBlob, blob.intf());
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, const IBPP::Array& array)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[Array]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[Array]", "The statement does not take parameters.");

	mInDescr->SetValue(param, ivArray, (void*)array.intf());
	mInMissing[param-1] = false;
}

void StatementImpl::Set(int param, const IBPP::DBKey& key)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[DBKey]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[DBKey]", "The statement does not take parameters.");

	mInDescr->SetValue(param, ivDBKey, (void*)&key);
	mInMissing[param-1] = false;
}

/*
void StatementImpl::Set(int param, const IBPP::Value& value)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Set[Value]", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Set[Value]", "The statement does not take parameters.");

	//mInDescr->SetValue(param, ivDBKey, (void*)&key);
	//mInMissing[param-1] = false;
}
*/

bool StatementImpl::IsNull(int column)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::IsNull", "The row is not initialized.");

	return mOutDescr->IsNull(column);
}

bool StatementImpl::Get(int column, bool* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == 0)
		throw ExceptionImpl("Statement::Get", "Null pointer detected");

	void* pvalue = mOutDescr->GetValue(column, ivBool);
	if (pvalue != 0)
		*retvalue = (*(short*)pvalue == 0 ? false : true);
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, bool& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* pvalue = mOutDescr->GetValue(column, ivBool);
	if (pvalue != 0)
		retvalue = (*(short*)pvalue == 0 ? false : true);
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, char* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == 0)
		throw ExceptionImpl("Statement::Get", "Null pointer detected");

	int sqllen;
	void* pvalue = mOutDescr->GetValue(column, ivByte, &sqllen);
	if (pvalue != 0)
	{
		memcpy(retvalue, pvalue, sqllen);
		retvalue[sqllen] = '\0';
	}
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, void* bindata, int& userlen)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (bindata == 0)
		throw ExceptionImpl("Statement::Get", "Null pointer detected");
	if (userlen < 0)
		throw ExceptionImpl("Statement::Get", "Length must be >= 0");

	int sqllen;
	void* pvalue = mOutDescr->GetValue(column, ivByte, &sqllen);
	if (pvalue != 0)
	{
		// userlen says how much bytes the user can accept
		// let's shorten it, if there is less bytes available
		if (sqllen < userlen) userlen = sqllen;
		memcpy(bindata, pvalue, userlen);
	}
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, std::string& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* pvalue = mOutDescr->GetValue(column, ivString);
	if (pvalue != 0)
		retvalue = *(std::string*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, short* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == 0)
		throw ExceptionImpl("Statement::Get", "Null pointer detected");

	void* pvalue = mOutDescr->GetValue(column, ivShort);
	if (pvalue != 0)
		*retvalue = *(short*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, short& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* pvalue = mOutDescr->GetValue(column, ivShort);
	if (pvalue != 0)
		retvalue = *(short*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, int* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == 0)
		throw ExceptionImpl("Statement::Get", "Null pointer detected");

	void* pvalue = mOutDescr->GetValue(column, ivLong);
	if (pvalue != 0)
		*retvalue = *(int*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, int& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* pvalue = mOutDescr->GetValue(column, ivLong);
	if (pvalue != 0)
		retvalue = *(int*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, long* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == 0)
		throw ExceptionImpl("Statement::Get", "Null pointer detected");

	void* pvalue = mOutDescr->GetValue(column, ivLong);
	if (pvalue != 0)
		*retvalue = *(long*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, long& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* pvalue = mOutDescr->GetValue(column, ivLong);
	if (pvalue != 0)
		retvalue = *(long*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, __int64* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == 0)
		throw ExceptionImpl("Statement::Get", "Null pointer detected");

	void* pvalue = mOutDescr->GetValue(column, ivLarge);
	if (pvalue != 0)
		*retvalue = *(__int64*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, __int64& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* pvalue = mOutDescr->GetValue(column, ivLarge);
	if (pvalue != 0)
		retvalue = *(__int64*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, float* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == 0)
		throw ExceptionImpl("Statement::Get", "Null pointer detected");

	void* pvalue = mOutDescr->GetValue(column, ivFloat);
	if (pvalue != 0)
		*retvalue = *(float*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, float& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* pvalue = mOutDescr->GetValue(column, ivFloat);
	if (pvalue != 0)
		retvalue = *(float*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, double* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");
	if (retvalue == 0)
		throw ExceptionImpl("Statement::Get", "Null pointer detected");

	void* pvalue = mOutDescr->GetValue(column, ivDouble);
	if (pvalue != 0)
		*retvalue = *(double*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, double& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* pvalue = mOutDescr->GetValue(column, ivDouble);
	if (pvalue != 0)
		retvalue = *(double*)pvalue;
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, IBPP::Timestamp& timestamp)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* pvalue = mOutDescr->GetValue(column, ivTimestamp, (void*)&timestamp);
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, IBPP::Date& date)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	if (mDatabase->Dialect() == 1)
	{
		// Dialect 1. IBPP::Date is supposed to work with old 'DATE'
		// fields which are actually ISC_TIMESTAMP.
		IBPP::Timestamp timestamp;
		void* pvalue = mOutDescr->GetValue(column, ivTimestamp, (void*)&timestamp);
		if (pvalue != 0) date = timestamp;
		return pvalue == 0 ? true : false;
	}
	else
	{
		void* pvalue = mOutDescr->GetValue(column, ivDate, (void*)&date);
		return pvalue == 0 ? true : false;
	}
}

bool StatementImpl::Get(int column, IBPP::Time& time)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* pvalue = mOutDescr->GetValue(column, ivTime, (void*)&time);
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, IBPP::Blob& retblob)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	/*
	retblob->AttachDatabase(mDatabase);
	retblob->AttachTransaction(mTransaction);
	*/
	void* pvalue = mOutDescr->GetValue(column, ivBlob, (void*)retblob.intf());
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, IBPP::DBKey& retkey)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* pvalue = mOutDescr->GetValue(column, ivDBKey, (void*)&retkey);
	return pvalue == 0 ? true : false;
}

bool StatementImpl::Get(int column, IBPP::Array& retarray)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	void* pvalue = mOutDescr->GetValue(column, ivArray, (void*)retarray.intf());
	return pvalue == 0 ? true : false;
}

/*
const IBPP::Value StatementImpl::Get(int column)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	//void* value = mOutDescr->GetValue(column, ivArray, (void*)retarray.intf());
	//return value == 0 ? true : false;
	return IBPP::Value();
}
*/

bool StatementImpl::IsNull(const std::string& name)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::IsNull", "The row is not initialized.");

	return IsNull(mOutDescr->GetVarNum(name));
}

bool StatementImpl::Get(const std::string& name, bool* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, bool& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, char* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get[char*]", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, void* retvalue, int& count)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get[void*,int]", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue, count);
}

bool StatementImpl::Get(const std::string& name, std::string& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::GetString", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, short* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, short& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, int* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, int& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, long* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, long& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, __int64* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, __int64& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, float* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, float& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, double* retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, double& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, IBPP::Timestamp& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, IBPP::Date& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, IBPP::Time& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string&name, IBPP::Blob& retblob)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retblob);
}

bool StatementImpl::Get(const std::string& name, IBPP::DBKey& retvalue)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retvalue);
}

bool StatementImpl::Get(const std::string& name, IBPP::Array& retarray)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name), retarray);
}

/*
const IBPP::Value StatementImpl::Get(const std::string& name)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Get", "The row is not initialized.");

	return Get(mOutDescr->GetVarNum(name));
}
*/

int StatementImpl::Columns(void)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Columns", "The row is not initialized.");

	return mOutDescr->DescrSize();
}

int StatementImpl::ColumnNum(const std::string& name)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::ColumnNum", "The row is not initialized.");
	if (name.empty())
		throw ExceptionImpl("Statement::ColumnNum", "0 parameter detected.");

	return mOutDescr->GetVarNum(name);
}

const char* StatementImpl::ColumnName(int varnum)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Columns", "The row is not initialized.");

	return mOutDescr->GetName(varnum);
}

const char* StatementImpl::ColumnAlias(int varnum)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Columns", "The row is not initialized.");

	return mOutDescr->GetAlias(varnum);
}

const char* StatementImpl::ColumnTable(int varnum)
{
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::Columns", "The row is not initialized.");

	return mOutDescr->GetTable(varnum);
}

IBPP::SDT StatementImpl::ColumnType(int varnum)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::ColumnType", "No statement has been prepared.");
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::ColumnType", "The statement does not return results.");

    return mOutDescr->GetType(varnum);
}

int StatementImpl::ColumnSize(int varnum)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::ColumnSize", "No statement has been prepared.");
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::ColumnSize", "The row is not initialized.");

	return mOutDescr->GetSize(varnum);
}

int StatementImpl::ColumnScale(int varnum)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::ColumnScale", "No statement has been prepared.");
	if (mOutDescr == 0)
		throw ExceptionImpl("Statement::ColumnScale", "The row is not initialized.");

	return mOutDescr->GetScale(varnum);
}

int StatementImpl::Parameters(void)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::Parameters", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::Parameters", "The statement uses no parameters.");

	return mInDescr->DescrSize();
}

IBPP::SDT StatementImpl::ParameterType(int varnum)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::ParameterType", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::ParameterType", "The statement uses no parameters.");

    return mInDescr->GetType(varnum);
}

int StatementImpl::ParameterSize(int varnum)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::ParameterSize", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::ParameterSize", "The statement uses no parameters.");

	return mInDescr->GetSize(varnum);
}

int StatementImpl::ParameterScale(int varnum)
{
	if (mHandle == 0)
		throw ExceptionImpl("Statement::ParameterScale", "No statement has been prepared.");
	if (mInDescr == 0)
		throw ExceptionImpl("Statement::ParameterScale", "The statement uses no parameters.");

	return mInDescr->GetScale(varnum);
}

IBPP::IStatement* StatementImpl::AddRef(void)
{
	ASSERTION(mRefCount >= 0);
	++mRefCount;

	return this;
}

void StatementImpl::Release(IBPP::IStatement*& Self)
{
	if (this != dynamic_cast<StatementImpl*>(Self))
		throw ExceptionImpl("Statement::Release", "Invalid Release()");

	ASSERTION(mRefCount >= 0);

	--mRefCount;

	if (mRefCount <= 0) delete this;
	Self = 0;
}

//	(((((((( OBJECT INTERNAL METHODS ))))))))

void StatementImpl::AttachDatabase(DatabaseImpl* database)
{
	if (database == 0)
		throw ExceptionImpl("Statement::AttachDatabase",
			"Can't attach a 0 IDatabase object.");

	if (mDatabase != 0) mDatabase->DetachStatement(this);
	mDatabase = database;
	mDatabase->AttachStatement(this);
}

void StatementImpl::DetachDatabase(void)
{
	if (mDatabase == 0)
		throw ExceptionImpl("Statement::DetachDatabase",
			"No IDatabase was attached.");

	try {Close();} catch(IBPP::Exception&) {}
	mDatabase->DetachStatement(this);
	mDatabase = 0;
}

void StatementImpl::AttachTransaction(TransactionImpl* transaction)
{
	if (transaction == 0)
		throw ExceptionImpl("Statement::AttachTransaction",
			"Can't attach a 0 ITransaction object.");

	if (mTransaction != 0) mTransaction->DetachStatement(this);
	mTransaction = transaction;
	mTransaction->AttachStatement(this);
}

void StatementImpl::DetachTransaction(void)
{
	if (mTransaction == 0)
		throw ExceptionImpl("Statement::DetachTransaction",
			"No ITransaction was attached.");

	try {Close();} catch(IBPP::Exception&) {}
	mTransaction->DetachStatement(this);
	mTransaction = 0;
}

void StatementImpl::CursorFree(void)
{
	if (mResultSetAvailable)
	{
		if (mHandle != 0)
		{
			IBS status;
			(*gds.Call()->m_dsql_free_statement)(status.Self(), &mHandle, DSQL_close);
			if (status.Errors())
				throw ExceptionImpl(status, "StatementImpl::CursorFree",
					"isc_dsql_free_statement failed.");
		}
		mResultSetAvailable = false;
	}
}

StatementImpl::StatementImpl(DatabaseImpl* database, TransactionImpl* transaction,
	const std::string& sql)
	: mRefCount(0), mHandle(0), mDatabase(0), mTransaction(0),
	mInDescr(0), mInMissing(0), mOutDescr(0),
	mResultSetAvailable(false), mType(IBPP::stUnknown)
{
	AttachDatabase(database);
	if (transaction != 0) AttachTransaction(transaction);
	if (! sql.empty()) Prepare(sql);
}

StatementImpl::~StatementImpl()
{
	try {Close();} catch (IBPP::Exception&) {}
	if (mTransaction != 0) mTransaction->DetachStatement(this);
	if (mDatabase != 0) mDatabase->DetachStatement(this);
}

//
//	EOF
//
